#!/usr/bin/env python

from distutils.core import setup

setup(
    name="aquatic",
    version="0.0.0",
    description="A quantization-aware training library for PyTorch",
    author="Sadique Sheik, Gregor Lenz",
    author_email="sadique.sheik@synsense.ai",
    packages=["aquatic"],
)
